/*
  This kernel fuses batch normalization, bias addition, division, and Swish activation.
  Each block processes one feature column. It uses vectorized loads (processing 4 elements per iteration)
  and warp-level reduction via __shfl_down_sync to compute the feature-wise sum and sum of squares efficiently.
  The computed mean and variance are then used to update running statistics and normalize the inputs.
*/

#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <vector>
#include <math.h>

// Kernel: fused_bn_swish_opt_kernel
// Each block handles one feature (column) from the x_linear tensor of shape [batch_size, out_features].
// We use vectorized loads and warp-level reduction to compute the feature-wise sum and sumsq.

template <typename scalar_t>
__global__ void fused_bn_swish_opt_kernel(
    const scalar_t* __restrict__ x_linear,       // Input: [batch_size, out_features]
    scalar_t* __restrict__ output,               // Output: same shape as x_linear
    const scalar_t* __restrict__ bn_weight,        // BatchNorm scale, shape: [out_features]
    const scalar_t* __restrict__ bn_bias,          // BatchNorm bias, shape: [out_features]
    scalar_t* __restrict__ bn_running_mean,        // Running mean, shape: [out_features]
    scalar_t* __restrict__ bn_running_var,         // Running variance, shape: [out_features]
    const scalar_t* __restrict__ add_bias,         // Additional bias (1-element tensor)
    const float bn_eps,
    const float bn_momentum,
    const float divide_value,
    const int batch_size,
    const int out_features) {

    // Each block processes one output feature
    const int f = blockIdx.x;
    if (f >= out_features) return;
    const int tid = threadIdx.x;

    // Use vectorized processing with a vector size of 4
    const int vec_size = 4;
    const int vec_limit = (batch_size / vec_size) * vec_size;

    // Each thread accumulates partial sums and sums-of-squares
    float thread_sum = 0.0f;
    float thread_sumsq = 0.0f;

    // Vectorized loop: process multiples of 4
    for (int i = tid * vec_size; i < vec_limit; i += blockDim.x * vec_size) {
        // Manually load 4 elements. Note: Elements are accessed with a stride of out_features.
        float v0 = static_cast<float>(x_linear[(i + 0) * out_features + f]);
        float v1 = static_cast<float>(x_linear[(i + 1) * out_features + f]);
        float v2 = static_cast<float>(x_linear[(i + 2) * out_features + f]);
        float v3 = static_cast<float>(x_linear[(i + 3) * out_features + f]);
        thread_sum   += v0 + v1 + v2 + v3;
        thread_sumsq += v0 * v0 + v1 * v1 + v2 * v2 + v3 * v3;
    }

    // Remainder loop: process remaining elements
    for (int i = vec_limit + tid; i < batch_size; i += blockDim.x) {
        float v = static_cast<float>(x_linear[i * out_features + f]);
        thread_sum   += v;
        thread_sumsq += v * v;
    }

    // Warp-level reduction using shuffle
    unsigned int mask = 0xffffffff;
    for (int offset = warpSize / 2; offset > 0; offset /= 2) {
        thread_sum   += __shfl_down_sync(mask, thread_sum, offset);
        thread_sumsq += __shfl_down_sync(mask, thread_sumsq, offset);
    }

    // Use shared memory to accumulate results from each warp
    extern __shared__ float shared[]; // shared memory: first part for sums, second for sumsq
    const int warpId = tid >> 5;         // tid / 32
    const int lane   = tid & 31;           // tid % 32
    const int numWarps = blockDim.x / 32;

    if (lane == 0) {
        shared[warpId] = thread_sum;
        shared[warpId + numWarps] = thread_sumsq;
    }
    __syncthreads();

    // Let the first warp reduce the per-warp partials
    float total_sum = 0.0f;
    float total_sumsq = 0.0f;
    if (tid < numWarps) {
        total_sum = shared[tid];
        total_sumsq = shared[tid + numWarps];
    }
    // Only first warp threads participate
    if (tid < 32) {
        for (int offset = numWarps / 2; offset > 0; offset /= 2) {
            if (tid < offset) {
                total_sum   += shared[tid + offset];
                total_sumsq += shared[tid + offset + numWarps];
            }
        }
    }

    // Store computed mean and variance in shared memory for use in second pass
    if (tid == 0) {
        float mean = total_sum / batch_size;
        float var = total_sumsq / batch_size - mean * mean;
        // Update running statistics (no atomics needed as one block per feature)
        bn_running_mean[f] = bn_running_mean[f] * (1.0f - bn_momentum) + mean * bn_momentum;
        bn_running_var[f]  = bn_running_var[f]  * (1.0f - bn_momentum) + var * bn_momentum;
        shared[0] = mean;
        shared[1] = var;
    }
    __syncthreads();

    // Load the computed mean and variance
    float mean = shared[0];
    float var  = shared[1];
    float inv_std = rsqrtf(var + bn_eps);
    float gamma = static_cast<float>(bn_weight[f]);
    float beta  = static_cast<float>(bn_bias[f]);
    float extra_bias = static_cast<float>(add_bias[0]);

    // Second pass: apply BN, add extra bias, divide and apply Swish activation
    // Process vectorized loop
    for (int i = tid * vec_size; i < vec_limit; i += blockDim.x * vec_size) {
        #pragma unroll
        for (int j = 0; j < vec_size; j++) {
            const int idx = (i + j) * out_features + f;
            float val = static_cast<float>(x_linear[idx]);
            float normalized = (val - mean) * inv_std;
            // Apply batch norm transform with bias and extra bias
            float transformed = fmaf(normalized, gamma, beta) + extra_bias;
            float divided = transformed / divide_value;
            float swish = divided / (1.f + expf(-divided));
            output[idx] = static_cast<scalar_t>(swish);
        }
    }
    
    // Process remaining elements
    for (int i = vec_limit + tid; i < batch_size; i += blockDim.x) {
        const int idx = i * out_features + f;
        float val = static_cast<float>(x_linear[idx]);
        float normalized = (val - mean) * inv_std;
        float transformed = fmaf(normalized, gamma, beta) + extra_bias;
        float divided = transformed / divide_value;
        float swish = divided / (1.f + expf(-divided));
        output[idx] = static_cast<scalar_t>(swish);
    }
}


// Host function to launch the kernel
torch::Tensor module_fn_cuda(
    torch::Tensor x,
    float bn_eps,
    float bn_momentum,
    float divide_value,
    torch::Tensor weight,
    torch::Tensor bias,
    torch::Tensor bn_weight,
    torch::Tensor bn_bias,
    torch::Tensor bn_running_mean,
    torch::Tensor bn_running_var,
    torch::Tensor add_bias) {

    const auto batch_size = x.size(0);
    const auto out_features = weight.size(0);

    // Ensure input tensors are contiguous
    x = x.contiguous();
    weight = weight.contiguous();
    bias = bias.contiguous();

    // Perform linear transformation: x_linear = x @ weight.T + bias
    auto x_linear = torch::addmm(bias, x, weight.t());
    auto output = torch::empty_like(x_linear);

    // Launch configuration: one block per feature column
    const int threads = 128;
    const int blocks = out_features;
    // Shared memory: allocate 2 floats per warp (numWarps = threads/32)
    const size_t shared_mem_size = 2 * (threads / 32) * sizeof(float);

    AT_DISPATCH_FLOATING_TYPES(x_linear.scalar_type(), "fused_bn_swish_opt_kernel", ([&] {
        fused_bn_swish_opt_kernel<scalar_t><<<blocks, threads, shared_mem_size>>>(
            x_linear.data_ptr<scalar_t>(),
            output.data_ptr<scalar_t>(),
            bn_weight.data_ptr<scalar_t>(),
            bn_bias.data_ptr<scalar_t>(),
            bn_running_mean.data_ptr<scalar_t>(),
            bn_running_var.data_ptr<scalar_t>(),
            add_bias.data_ptr<scalar_t>(),
            bn_eps,
            bn_momentum,
            divide_value,
            batch_size,
            out_features);
    }));

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn_cuda, "Fused BN, bias add, division and Swish forward (CUDA optimized)");
}
